# RPT_AR
If you fail to install and run this tracker, please email us 
(andyzhangchunhui@gmail.com, zhangkangkai@iie.ac.cn)

# Requirements
## Hardware Requirements
* GPU: NVIDIA TITAN Xp (8GB Memory)
* CPU: Intel® Core™ E5-2630 CPU @ 2.40GHz 
* SSD: Intel Disk 1920G
* RAM: 192GB

## Software Dependency
* OS: Ubuntu 18.04 LTS
* Anaconda 3
* CUDA 10.0, CUDNN 7.6.0
* Python libraries: Python 3.6, Pytorch 1.1.0

Please note that other versions of Pytorch may have some problems

## Required packages for python 3.6

    pytorch == 1.1.0
    
    numpy==1.17.3
    
    opencv-python==4.1.1.26
    
    torchvision == 0.2.0
    
    yacs == 0.1.6
    
    tqdm == 4.32.0
    
    pyyaml == 5.3
    
    matplotlib == 3.1.1
       
    cython == 0.29.14
    
    pandas == 0.25.3
    
    visdom == 0.1.8.9

## Compiled Libraries
    DeformConv
   
# Installation
This document contains detailed instructions for installing the necessary dependencies for **RPT_AR**. 
The instructions have been tested on an Ubuntu 18.04 system.  

Install the environment by executing the script `install.sh`.     
ATTENTION   we modified vot.py ./tools/vot.py, so it can output the offset of target mask.
    
Put the pretrained checkpoint of mask branch [SegmNet.pth.tar](https://drive.google.com/file/d/19KogscO16O0ASch6IS81Ebr7gwX5B2Ah/view?usp=sharing) into subfolder `'./pytracking/networks/SegmNet.pth.tar'`.

Put the pretrained checkpoint of location branch [siamreppoints.model](https://drive.google.com/file/d/12m2UaO9NtLE8mK532h7CJ3P-H482tGXO/view?usp=sharing) into the subfolder `'./tools/siamreppoints.model'`. 

Above pretrained checkpoints can also been downloaded from [VOT2020](http://data.votchallenge.net/vot2020/trackers/RPT-code-2020-05-03T14_38_51.902820.zip).

### RPT_AR-Related Part
* Create and activate a conda environment
```bash
conda create -n VOT2021 python=3.6
conda activate VOT2021
```

* Install PyTorch 
```bash
conda install pytorch==1.1.0 torchvision==0.2.0 cudatoolkit=10.0 -c pytorch
```

* Install matplotlib, pandas, tqdm, opencv, scikit-image, visdom, easydict and tikzplotlib 
```bash
conda install matplotlib pandas tqdm
pip install opencv-python visdom scikit-image easydict tikzplotlib
```


* Install ninja-build for Precise ROI pooling 
```bash
sudo apt-get install ninja-build
```
In case of issues, we refer to https://github.com/vacancy/PreciseRoIPooling. 


* Install jpeg4py 
```bash
pip install jpeg4py 
```


* Setup the environment 
Create the default environment setting files. 
```bash
# Change directory to <PATH_of_RPT_AR>
# e.g., /home/zhangchunhui/VOT2021Challenge/RPT_AR  
cd /home/zhangchunhui/VOT2021Challenge/RPT_AR

# Environment settings for pytracking. Saved at pytracking/evaluation/local.py
python -c "from pytracking.evaluation.environment import create_default_local_file; create_default_local_file()"

# Environment settings for ltr. Saved at ltr/admin/local.py
python -c "from ltr.admin.environment import create_default_local_file; create_default_local_file()"
```

You can modify these files to set the paths to datasets, results paths, etc.  
* Add the project path to environment variables  
Open ~/.bashrc, and add the following line to the end. Note to change <PATH_of_RPT_AR> to your real path.
```
export PYTHONPATH=<PATH_of_RPT_AR>:$PYTHONPATH
```

* Download the pre-trained network 
[ARnet_seg_mask_ep0040.pth.tar](https://drive.google.com/file/d/1bRvaSjYksmIval_HKbe_dvtzUq9zDGSv/view?usp=sharing) for Alpha-Refine 
and put it under the subfolder `'./pytracking/checkpoints/ltr/ARcm_seg/ARcm_coco_seg_only_mask_384/'`.

### VOT-Toolkit-Related Part
* Install vot-toolkit in the current conda environment
```bash
pip install git+https://github.com/votchallenge/vot-toolkit-python
```
* Reinstall pillow to avoid "ImportError: cannot import name 'PILLOW_VERSION' from 'PIL'"
```
conda install pillow=6.1
```
* Modify ['paths' in 'trackers_RPT_AR.ini'](/home/zhangchunhui/VOT2021Challenge/RPT_AR/vot-workspace/trackers_RPT_AR.ini) to the real path on your machine. 

# Evaluation
* Go to RPT_AR dir and modify some paths  
Modify paths in 'trackers_RPT_AR.ini' and 'run.sh' to your real paths.

* Run RPT_AR and analyze results  
**Note**: After download VOT2021-ST dataset, please move it to SSD for improving I/O speed :)  
You can create a soft link called 'sequences' and put it under '/home/zhangchunhui/VOT2021Challenge/RPT_AR/vot-workspace' dir.

* In vot-workspace (e.g., /home/zhangchunhui/VOT2021Challenge/RPT_AR/vot-workspace), run the command:
```
bash run.sh
```
