from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import vot
from vot import VOT
import sys
import time
import cv2
import numpy as np
import torch
import argparse
import os


filepath = os.path.abspath(__file__)
filelocation = os.path.split(filepath)[0]
rootpath = os.path.dirname(filelocation) 
sys.path.append(rootpath)

os.environ['CUDA_VISIBLE_DEVICES'] = '1'

from siamreppoints.core.config import cfg
from siamreppoints.models.model_builder import ModelBuilder
from siamreppoints.tracker.tracker_builder import build_tracker
from siamreppoints.utils.bbox import get_axis_aligned_bbox
from siamreppoints.utils.model_load import load_pretrain
torch.set_num_threads(1)
from pytracking.evaluation import Tracker as Trackers

#######################
'''ARcm_seg model'''
from pytracking_AR.ARcm_seg import ARcm_seg
'''other utils'''
from pytracking_AR.vot20_utils import *
#######################

parser = argparse.ArgumentParser(description='SiamRP tracker')

parser.add_argument('--config', default=os.path.join(rootpath,'experiments/siamreppoints/config.yaml'),type=str, help='config file')
parser.add_argument('--snapshot', default=os.path.join(rootpath,'tools/snapshot/siamreppoints.model'), type=str, help='tracker model name')
parser.add_argument('--d3smodel', default=os.path.join(rootpath,'pytracking/networks/SegmNet.pth.tar'), type=str, help='segmentation model name')
parser.add_argument('--video_name', default='', type=str, help='videos or image files')
args = parser.parse_args()

class NCCTracker(object):

    # def __init__(self, image, mask):
    #######################
    def __init__(self, image, mask, refine_model_name='ARcm_coco_seg_only_mask_384', threshold=0.15):
        self.THRES = threshold
    ######################
        # generate bbox as vot2019
        rotated_bbox = self._mask_post_processing(mask)
        rotated_bbox=np.array([rotated_bbox[0][0],rotated_bbox[0][1],rotated_bbox[1][0],rotated_bbox[1][1],rotated_bbox[2][0],rotated_bbox[2][1],rotated_bbox[3][0],rotated_bbox[3][1]])
        cx,cy,w,h = self.get_axis_aligned_bbox(rotated_bbox)
        gt_bbox_ = [cx-w/2, cy-h/2, w, h]

        # load config        
        cfg.merge_from_file(args.config)
        
        # create model
        model = ModelBuilder()
    
        # load model
        model = load_pretrain(model, args.snapshot).cuda().eval()

        # build tracker
        self.tracker = build_tracker(model)
        
        # build segmentation net
        tracker_name='segm'
        tracker_param='default_params'
        run_id=None
        d3strackers = [Trackers(tracker_name, tracker_param, run_id)]#tracker_name tracker_param,run_id
        params=d3strackers[0].get_parameters()
        params.segm_net_path=args.d3smodel
        self.d3stracker=d3strackers[0].tracker_class(params)
        
        # init SiamRP 
        self.tracker.init(image, gt_bbox_)
        d3simg=cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        self.d3stracker.initialize(d3simg,gt_bbox_, mask*255)

        ####################################
        # RectFusing = self.tracker.rect_fusing(gt_bbox_)
        ####################################

        ####################################
        '''Alpha-Refine'''
        refine_root = os.path.join('/home/zhangchunhui/VOT2021Challenge/RPT_AR/pytracking/checkpoints/ltr/ARcm_seg')
        refine_path = os.path.join(refine_root, refine_model_name)
        '''input size: 384x384'''
        self.alpha = ARcm_seg(refine_path, input_sz=384)
        ####################################

    #############################
    def initialize(self, image, mask):
        img_RGB = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        region = rect_from_mask(mask)
        self.H, self.W, _ = img_RGB.shape
        gt_bbox_np = np.array(region).astype(np.float32)
        '''initilize refinement module for specific video'''
        self.alpha.initialize(img_RGB, np.array(gt_bbox_np))
    ##############################

    def track(self, image):
        '''base tracker'''
        outputs = self.tracker.track(image)
        pred_bbox = outputs['bbox']
        score = outputs['best_score']
        
        d3simg=cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        outputs=self.d3stracker.loop_track(d3simg, outputs['bbox'])
        
        # make sure no empty mask output
        mask = (outputs['mask']>0.3).astype(np.uint8)
        # if np.where(mask > 0)[0].shape[0] > 0:
        #     return mask, score
        # else:
        #     img = np.zeros(mask.shape, dtype=np.uint8)
        #     img[int(pred_bbox[1]):int(pred_bbox[1]+pred_bbox[3]), int(pred_bbox[0]):int(pred_bbox[0]+pred_bbox[2])] = 1
        #     mask = img
        # return mask, score

        ############################################
        # '''Step2: Mask report'''
        img_RGB = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        bbox_new = pred_bbox
        pred_mask, search, search_mask = self.alpha.get_mask(img_RGB, np.array(bbox_new),vis=True)
        final_mask = (pred_mask > self.THRES).astype(np.uint8)
        search_region = search.astype(np.uint8)
        search_mask = (search_mask > self.THRES).astype(np.uint8)
        # return final_mask, score, bbox_new, final_mask, search_region, search_mask

        ############################################
        # mask fusing
        MM = ((mask + final_mask) > 1).astype(np.uint8)
        if np.where(MM > 0)[0].shape[0] > 0:
            # rect fusing
            # NewRect = rect_from_mask(MM)
            # RectFusing = self.tracker.rect_fusing(NewRect)
            # print(pred_bbox, NewRect, RectFusing)
            # return mask, score, bbox_new, final_mask, search_region, search_mask, MM, NewRect
            return mask, score, bbox_new, final_mask, search_region, search_mask, MM
        else:
            img = np.zeros(mask.shape, dtype=np.uint8)
            img[int(pred_bbox[1]):int(pred_bbox[1]+pred_bbox[3]), int(pred_bbox[0]):int(pred_bbox[0]+pred_bbox[2])] = 1
            mask = img
            # NewRect = pred_bbox
        # return mask, score, bbox_new, final_mask, search_region, search_mask, MM, NewRect
        return mask, score, bbox_new, final_mask, search_region, search_mask, MM
        ############################################

    def _mask_post_processing(self, mask):
        target_mask = (mask > 0.5)
        target_mask = target_mask.astype(np.uint8)
        if cv2.__version__[-5] == '4':
            contours, _ = cv2.findContours(target_mask, 
                    cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        else:
            _, contours, _ = cv2.findContours(target_mask, 
                    cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        cnt_area = [cv2.contourArea(cnt) for cnt in contours]
        if len(contours) != 0 and np.max(cnt_area) > 50:#cnt_area=100
            contour = contours[np.argmax(cnt_area)] 
            polygon = contour.reshape(-1, 2)

            ## the following code estimate the shape angle with ellipse
            ## then fit a axis-aligned bounding box on the rotated image
            
            ellipseBox = cv2.fitEllipse(polygon)
            # get the center of the ellipse and the angle
            angle = ellipseBox[-1]
            #print(angle)
            center = np.array(ellipseBox[0])
            axes = np.array(ellipseBox[1])
            
            # get the ellipse box
            ellipseBox = cv2.boxPoints(ellipseBox)
            
            #compute the rotation matrix
            rot_mat = cv2.getRotationMatrix2D((center[0],center[1]), angle, 1.0)
            
            # rotate the ellipse box
            one = np.ones([ellipseBox.shape[0],3,1])
            one[:,:2,:] = ellipseBox.reshape(-1,2,1)
            ellipseBox = np.matmul(rot_mat, one).reshape(-1,2)
            
            # to xmin ymin xmax ymax
            xs = ellipseBox[:,0]
            xmin, xmax = np.min(xs), np.max(xs)
            ys = ellipseBox[:,1]
            ymin, ymax = np.min(ys), np.max(ys)
            ellipseBox = [xmin, ymin, xmax, ymax]
            
            # rotate the contour
            one = np.ones([polygon.shape[0],3,1])
            one[:,:2,:] = polygon.reshape(-1,2,1)
            polygon = np.matmul(rot_mat, one).astype(int).reshape(-1,2)
            
            # remove points outside of the ellipseBox
            logi = polygon[:,0]<=xmax
            logi = np.logical_and(polygon[:,0]>=xmin, logi)
            logi = np.logical_and(polygon[:,1]>=ymin, logi)
            logi = np.logical_and(polygon[:,1]<=ymax, logi)
            polygon = polygon[logi,:]
            
            x,y,w,h = cv2.boundingRect(polygon)
            bRect = [x, y, x+w, y+h]
            
            # get the intersection of ellipse box and the rotated box
            x1, y1, x2, y2 = ellipseBox[0], ellipseBox[1], ellipseBox[2], ellipseBox[3]
            tx1, ty1, tx2, ty2 = bRect[0], bRect[1], bRect[2], bRect[3]
            xx1 = min(max(tx1, x1, 0), target_mask.shape[1]-1)
            yy1 = min(max(ty1, y1, 0), target_mask.shape[0]-1)
            xx2 = max(min(tx2, x2, target_mask.shape[1]-1), 0)
            yy2 = max(min(ty2, y2, target_mask.shape[0]-1), 0)
            
            rotated_mask = cv2.warpAffine(target_mask, rot_mat,(target_mask.shape[1],target_mask.shape[0]))
            
            # refinement
            alpha_factor = 0.2583 # cfg.TRACK.FACTOR
            while True:
                if np.sum(rotated_mask[int(yy1):int(yy2),int(xx1)]) < (yy2-yy1)*alpha_factor:
                    temp = xx1+(xx2-xx1)*0.02
                    if not (temp >= target_mask.shape[1]-1 or xx2-xx1 < 1):
                        xx1 = temp
                    else:
                        break
                else:
                    break
            while True:
                if np.sum(rotated_mask[int(yy1):int(yy2),int(xx2)]) < (yy2-yy1)*alpha_factor:
                    temp = xx2-(xx2-xx1)*0.02
                    if not (temp <= 0 or xx2-xx1 < 1):
                        xx2 = temp
                    else:
                        break
                else:
                    break
            while True:
                if np.sum(rotated_mask[int(yy1),int(xx1):int(xx2)]) < (xx2-xx1)*alpha_factor:
                    temp = yy1+(yy2-yy1)*0.02
                    if not (temp >= target_mask.shape[0]-1 or yy2-yy1 < 1):
                        yy1 = temp
                    else:
                        break
                else:
                    break
            while True:
                if np.sum(rotated_mask[int(yy2),int(xx1):int(xx2)]) < (xx2-xx1)*alpha_factor:
                    temp = yy2-(yy2-yy1)*0.02
                    if not (temp <= 0 or yy2-yy1 < 1):
                        yy2 = temp
                    else:
                        break
                else:
                    break
            
            prbox = np.array([[xx1,yy1],[xx2,yy1],[xx2,yy2],[xx1,yy2]])
            
            # inverse of the rotation matrix
            M_inv = cv2.invertAffineTransform(rot_mat)
            # project the points back to image coordinate
            one = np.ones([prbox.shape[0],3,1])
            one[:,:2,:] = prbox.reshape(-1,2,1)
            prbox = np.matmul(M_inv, one).reshape(-1,2)
            
            rbox_in_img = prbox
        else:  # empty mask
            # location = cxy_wh_2_rect(self.center_pos, self.size)
            location = [0,0,1,1]
            rbox_in_img = np.array([[location[0], location[1]],
                        [location[0] + location[2], location[1]],
                        [location[0] + location[2], location[1] + location[3]],
                        [location[0], location[1] + location[3]]])
        return rbox_in_img
    def get_axis_aligned_bbox(self,region):
        """ convert region to (cx, cy, w, h) that represent by axis aligned box
        """
        nv = region.size
        if nv == 8:
            cx = np.mean(region[0::2])
            cy = np.mean(region[1::2])
            x1 = min(region[0::2])
            x2 = max(region[0::2])
            y1 = min(region[1::2])
            y2 = max(region[1::2])
            A1 = np.linalg.norm(region[0:2] - region[2:4]) * \
                np.linalg.norm(region[2:4] - region[4:6])
            A2 = (x2 - x1) * (y2 - y1)
            s = np.sqrt(A1 / A2)
            w = s * (x2 - x1) + 1
            h = s * (y2 - y1) + 1
        else:
            x = region[0]
            y = region[1]
            w = region[2]
            h = region[3]
            cx = x+w/2
            cy = y+h/2
        return cx, cy, w, h         
    def _rect_from_mask(self, mask):
        '''
        create an axis-aligned rectangle from a given binary mask
        mask in created as a minimal rectangle containing all non-zero pixels
        '''
        # print(mask)
        x_ = np.sum(mask, axis=0)
        y_ = np.sum(mask, axis=1)
        x0 = np.min(np.nonzero(x_))
        x1 = np.max(np.nonzero(x_))
        y0 = np.min(np.nonzero(y_))
        y1 = np.max(np.nonzero(y_))
        return [x0, y0, x1 - x0 + 1, y1 - y0 + 1]

def make_full_size(x, offset, output_sz):
    '''
    zero-pad input x (right and down) to match output_sz
    x: numpy array e.g., binary mask
    output_sz: size of the output [width, height]
    '''
    if x.shape[0] == output_sz[1] and x.shape[1] == output_sz[0]:
        return x
    pad_x = output_sz[0] - int(x.shape[1]+offset[0])
    if pad_x < 0:
        x = x[:, :x.shape[1] + pad_x]
        # padding has to be set to zero, otherwise pad function fails
        pad_x = 0
        
    front_pad_x=output_sz[0]-pad_x-x.shape[1]
    if front_pad_x<0:
        x = x[:, x.shape[1] + front_pad_x:]
        front_pad_x=0
        
    pad_y = output_sz[1] - int(x.shape[0]+offset[1])
    if pad_y < 0:
        x = x[:x.shape[0] + pad_y, :]
        # padding has to be set to zero, otherwise pad function fails
        pad_y = 0
    front_pad_y=output_sz[1]-pad_y-x.shape[0]
    if front_pad_y<0:
        x = x[x.shape[0] + front_pad_y:, :]
        front_pad_y=0
    # print(front_pad_y+pad_y+x.shape[0],output_sz[1])
    # print(front_pad_x+pad_x+x.shape[1],output_sz[0])
    assert((front_pad_y+pad_y+x.shape[0])==output_sz[1])
    assert((front_pad_x+pad_x+x.shape[1])==output_sz[0])    
    return np.pad(x, ((front_pad_y, pad_y), (front_pad_x, pad_x)), 'constant', constant_values=0)

############################################
VIS  =False
save_root = os.path.join('vot2021_visualization')
if VIS and (not os.path.exists(save_root)):
    os.mkdir(save_root)
############################################

# handle = vot.VOT("mask")
handle = VOT("mask")
selection, offset = handle.region()

imagefile = handle.frame()
if not imagefile:
    sys.exit(0)

#################################
if VIS:
    '''for vis'''
    seq_name = imagefile.split('/')[-3]
    save_v_dir = os.path.join(save_root,seq_name)
    if not os.path.exists(save_v_dir):
        os.mkdir(save_v_dir)
    cur_time = int(time.time() % 10000)
    save_dir = os.path.join(save_v_dir, str(cur_time))
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
##################################

image = cv2.imread(imagefile)
# mask given by the toolkit ends with the target (zero-padding to the right and down is needed)
mask = make_full_size(selection, offset,(image.shape[1], image.shape[0]))
# print(mask)
# print(mask.shape)
if selection.max()==0:
    raise('selection.max() is 0')
if mask.max()==0:
    raise('mask.max()==0')
       
tracker = NCCTracker(image, mask)

##########################################
image_RGB = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)  # Right
mask_AR = make_full_size_AR(selection, (image.shape[1], image.shape[0]))
tracker.initialize(image_RGB, mask_AR)
##########################################

while True:
    imagefile = handle.frame()
    if not imagefile:
        break
    image = cv2.imread(imagefile)
    # m, confidence = tracker.track(image)
    # handle.report(m, confidence)
    #######################################
    # m, confidence, bbox_new, final_mask, search_region, search_mask, MM, NewRect = tracker.track(image)
    m, confidence, b1, final_mask, search, search_m, MM = tracker.track(image)
    # tracker.rect_fusing(NewRect)
    handle.report(MM, confidence)
    #######################################

    ##########################################################
    if VIS:
        '''Visualization'''
        # original image
        image_ori = image[:, :, ::-1].copy()  # RGB --> BGR
        image_name = imagefile.split('/')[-1]
        save_path = os.path.join(save_dir, image_name)
        cv2.imwrite(save_path, image_ori)
        # SiamRP box
        image_b = image_ori.copy()
        cv2.rectangle(image_b, (int(b1[0]), int(b1[1])),
                      (int(b1[0] + b1[2]), int(b1[1] + b1[3])), (0, 0, 255), 2)
        image_b_name = image_name.replace('.jpg', '_bbox.jpg')
        save_path = os.path.join(save_dir, image_b_name)
        cv2.imwrite(save_path, image_b)
        # search region
        search_bgr = search[:, :, ::-1].copy()
        search_name = image_name.replace('.jpg', '_search.jpg')
        save_path = os.path.join(save_dir, search_name)
        cv2.imwrite(save_path, search_bgr)
        # search region mask
        search_bgr_m = search_bgr.astype(np.float32)
        search_bgr_m[:, :, 1] += 127.0 * search_m
        search_bgr_m[:, :, 2] += 127.0 * search_m
        contours, _ = cv2.findContours(search_m, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        search_bgr_m = cv2.drawContours(search_bgr_m, contours, -1, (0, 255, 255), 4)
        search_bgr_m = search_bgr_m.clip(0, 255).astype(np.uint8)
        search_name_m = image_name.replace('.jpg', '_search_mask.jpg')
        save_path = os.path.join(save_dir, search_name_m)
        cv2.imwrite(save_path, search_bgr_m)
        # original image + mask
        image_m = image_ori.copy().astype(np.float32)
        image_m[:, :, 1] += 127.0 * m
        image_m[:, :, 2] += 127.0 * m
        contours, _ = cv2.findContours(m, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        image_m = cv2.drawContours(image_m, contours, -1, (0, 255, 255), 2)
        image_m = image_m.clip(0, 255).astype(np.uint8)
        image_mask_name_m = image_name.replace('.jpg', '_mask.jpg')
        save_path = os.path.join(save_dir, image_mask_name_m)
        cv2.imwrite(save_path, image_m)
        ##########################################################